### Richard G Brown
### GNU Makefile for CompressionNode and DecompressionNode
###
### Options
###
###	[blank]	build code
###	clean	remove all non-source files
###	build	build code
###	install	copy lils and dlls into place
###	all	clean build install
###
###	$(MQSIDIR) Base install directory of MQSI

MQSIDIR=/opt/mqsi
CFLAGS=-xCC 
LDFLAGS=

compressionNode: CompressorCatalogue.cat rgbnode.o librgbnodeutils.so libcompression.so
	cc $(CFLAGS) -G -o ./compressionNode.lil ./rgbnode.o -lz -L. -lrgbnodeutils -lcompression -L$(MQSIDIR)/lib -limbdfplg

rgbnode.o: rgbnode.c rgbnode.h rgbnodeutils.h compression.h rgbos.h CompressorCatalogue.cat
	cc $(CFLAGS) -mt -I. -I $(MQSIDIR)/include -I $(MQSIDIR)/include -I $(MQSIDIR)/include/plugin -c rgbnode.c -o rgbnode.o

rgbnodeutils.o: rgbnodeutils.c rgbnodeutils.h rgbos.h CompressorCatalogue.cat
	cc $(CFLAGS) -mt -I. -I $(MQSIDIR)/include -I $(MQSIDIR)/include -I $(MQSIDIR)/include/plugin -c rgbnodeutils.c -o rgbnodeutils.o

librgbnodeutils.so: rgbnodeutils.o
	ld $(LNFLAGS) -G -o librgbnodeutils.so rgbnodeutils.o -l imbdfplg -lc

compression.o: compression.h compression.c rgbnodeutils.h rgbos.h CompressorCatalogue.cat
	cc $(CFLAGS) -mt -I. -I $(MQSIDIR)/include -I $(MQSIDIR)/include -I $(MQSIDIR)/include/plugin -c compression.c -o compression.o

libcompression.so: compression.o librgbnodeutils.so
	ld $(LNFLAGS) -G -o libcompression.so compression.o -L. -lrgbnodeutils

CompressorCatalogue.cat: compressorcat.mc
	perl process1.pl < compressorcat.mc > CompressionNode.msg
	perl process2.pl < compressorcat.mc > compressorcat.h
	gencat CompressorCatalogue.cat CompressionNode.msg

clean:
	rm -f *.o
	rm -f libcompression.so
	rm -f librgbnodeutils.so
	rm -f CompressorCatalogue.cat
	rm -f compressionNode.lil
	rm -f compressorcat.h
	rm -f CompressionNode.msg

install:
	cp ./librgbnodeutils.so $(MQSIDIR)/lib
	cp ./libcompression.so $(MQSIDIR)/lib
	cp ./compressionNode.lil $(MQSIDIR)/lil
	cp ./CompressorCatalogue.cat $(MQSIDIR)/messages/ 
	cp ./CompressorCatalogue.cat $(MQSIDIR)/messages/En_US/
	rm -f /usr/lib/locale/C/LC_MESSAGES/CompressorCatalogue.cat
	ln -s $(MQSIDIR)/messages/En_US/CompressorCatalogue.cat /usr/lib/locale/C/LC_MESSAGES/CompressorCatalogue.cat

all: clean compressionNode install

build: compressionNode
