/*  Compression / Decompression plug-in */
/*  nodes for MQSeries Integrator version 2 */

/*  Some functions based on MQSI sample plug-in code */
/*  Some functions based on code written by Neil Kolban, with permission */

/*  Helper utilities */
/*  Richard G Brown */

#define __RGB_EXPORT 1
#include "rgbnodeutils.h"

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX messageSourcePointer=0;

/* //////////////////////////////////////////////////////////////////// */
/*  String handling functions derived from sample plug in code          */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX CciCharCpy(CciChar* target, const CciChar* source)
{
  static char* functionName = (char*)"CciCharCpy()";

  CciChar c ;

  compLog(1,functionName);

  compLog(0, "Copying %s to memory location %p", DmbString(source),
	  (void*)target);
  
  for (;;)
    {
      c = *source ;
      source += 1 ;
      *target = c ;
      target += 1 ;
      if ( 0 == c ) break ;
    }

  compLog(-1,functionName);
  return;
}

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX CciCharNCpy(CciChar* target, const CciChar* source, 
		 int size)
{
  static char* functionName = (char*)"CciCharNCpy()";

  CciChar c ;
  int i;

  compLog(1,functionName);
  compLog(0, "Copying %s to memory location %p using %d CciChars", 
	  DmbString(source),
	  (void*)target,
	  size);

  /*  leave space at end for null */
  for (i=0;i<size-1;i+=1) 
    {
      c = *source ;
      *target = c ;
      target += 1 ;
      if ( 0 != c ) source += 1 ;
    }

  /*  write zero into last space in case we truncated the source */
  *target=0;

  compLog(-1,functionName);

  return;
}

int __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX CciCharLen(const CciChar* s)
{
  /*  returns length of string, not including the terminal null */
  static char* functionName = (char*)"CciCharLen()";

  int i=0 ;

  compLog(1,functionName);

  /* compLog(0, "Determining length of %s.", DmbString(s));*/
  /* It is not wise to call DmbString here since its implementation depends
     on CciCharLen!! */

  compLog(0, "Determining length of input string: (cannot display to avoid circularity)");

  while (*s != 0 )
    {
      s += 1 ;
      i += 1 ;
    }

  compLog(0, "Returning %d", i);

  compLog(-1,functionName);
  return i ;
}

int __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX CciCharCmp(const CciChar * s1, const CciChar* s2)
{
  /*  returns zero only if strings are equal */
  static char* functionName = (char*)"CciCharCmp()";

  CciChar c1 ;
  CciChar c2 ;

  compLog(1,functionName);

  compLog(0, "Comparing strings %s and %s", DmbString(s1),
	  DmbString(s2));

  for (;;)
    {
      c1 = *s1 ;
      s1 += 1 ;
      c2 = *s2 ;
      s2 += 1 ;
      if ( (0 == c1) || (c1 != c2) ) break ;
    }

  if (c1-c2==0)
    compLog(0, "They are the same");
  else
    compLog(0, "They differ");

  compLog(-1,functionName);

  return (c1-c2) ;
}

CciChar __RGB_IMPORTEXPORTPREFIX *  __RGB_IMPORTEXPORTSUFFIX CciString(const char* source)
{
  /*  The returned string must be freed when finished with */

  static char* functionName = (char*)"CciString()";

  int maxChars;
  int rc ;
  CciChar* buffer; 

  compLog(1,functionName);

  compLog(0, "source string: %s", source);

  maxChars = strlen(source) + 1 ;

  compLog(0, "Allocating %d bytes.", maxChars);
  buffer = (CciChar*)malloc(maxChars * sizeof(CciChar)) ;
  TESTMALLOC(buffer);

  /* I am not sure what we can assume about cciMbsToUcs so let's
     set this buffer to zero so we have a null terminated
     string even if the function does not provide one */

  compLog(0, "Setting buffer to nulls");
  memset(buffer, '\0', maxChars * sizeof(CciChar));

  /* The following comment is from the sample code...
     I'm not sure this is a terribly wonderful way to 
     proceed but it will do for now... */
  
  /* We don't know the compiler's code page at this stage, 
     but 1208 (utf8 Unicode) will generally be appropriate 
     for ASCII systems */

  compLog(0, "Calling cciMbsToUcs");
  cciMbsToUcs(&rc, source, buffer, maxChars, 1208) ;
  compLog(0, "Return code was %d.", rc);

  compLog(-1,functionName);

  return buffer ;
}

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX CciNString(const char* source, const unsigned int size)
{
  /* Maximum number of characters in Unicode representation */
  /* The bytes may not be NULL terminated so we need to make a copy */

  static char* functionName = (char*)"CciNString()";

  int bufferSize;
  int rc;
  CciChar* buffer;
  char* inputBuffer; /* To create null terminated string */

  compLog(1,functionName);
  compLog(0, "Creating CciChar representation of %s using %d CciChars",
	  source, size+1);

  bufferSize = ((size + 1) * sizeof(CciChar));

  compLog(0, "Creating output buffer");
  buffer = (CciChar*)malloc(bufferSize);
  TESTMALLOC(buffer);

  compLog(0, "Creating buffer to hold truncated input string of size %d",
	  size+1);
  inputBuffer = (char*)malloc(size+1);
  TESTMALLOC(inputBuffer);

  compLog(0, "Copying truncated string into temporary char* buffer");
  memset(inputBuffer, '\0', size+1);
  memcpy(inputBuffer, source,size);

  compLog(0, "Setting output buffer to nulls");
  memset(buffer, '\0', bufferSize);

  /* We don't know the compiler's code page at this stage, 
     but 1208 (utf8 Unicode) will generally be appropriate 
     for ASCII systems (so says the sample code) */
  compLog(0, "Calling cciMbsToUcs");
  cciMbsToUcs(&rc, inputBuffer, buffer, size, 1208) ;
  compLog(0, "Return code was %d", rc);
  compLog(0, "Freeing temporary buffer");
  free(inputBuffer);

  compLog(-1,functionName);

  return buffer ;
}

char __RGB_IMPORTEXPORTPREFIX* __RGB_IMPORTEXPORTSUFFIX DmbString(const CciChar* source)
{
  /* We need a debug wrapper since the compLog() debug statements
     do not call free on the anonymous strings they produce.
     Therefore, we cannot allow them to be evaluated in a 
     production environment */
#ifdef _DEBUG
  return mbString(source);
#else
  return NULL;
#endif
}

char __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX mbString(const CciChar* source)
{
  static char* functionName = (char*)"mbString()";

  /* This function should not be called by functions which
     are not able to free the storage afterwards: i.e.
     debugging code that makes gratuitous use of anonymous
     dynamically generated strings should call DmbString 
     as a wrapper to this code so that non-debug versions 
     simply don't call this function at all... I don't care
     about debug builds leaking memory */

  int charCount;
  int maxBytes;
  char * buffer;
  int rc ;

  compLog(1,functionName);
  charCount = CciCharLen(source) ;
  maxBytes = (charCount+1) * 9 ;
  compLog(0, "Allocating %d bytes for output buffer.", maxBytes);

  buffer = (char *)malloc(maxBytes) ;
  TESTMALLOC(buffer);
  compLog(0, "Calling cciUcsToMbs");
  cciUcsToMbs(&rc, source, buffer, maxBytes, 1208) ;
  compLog(0, "Return code was %d", rc);
  compLog(-1,functionName);
  return buffer ;
}

/* //////////////////////////////////////////////////////////////////// */
/*  Context Handling */

/*  The general approach to node attributes taken is to store them
    all in a linked list containing key, value and next pointers.
    In this way, it is trivial to add extra attributes */

/*  get attribute */

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX getContextAttribute(CciChar* key, 
			     struct ContextAttribute* contextAttr)
{
  static char* functionName = "getContextAttribute()";
  CciChar* returnvalue = NULL;
  int iRc;
  
  compLog(1,functionName);

  if (contextAttr==NULL)
    {
      cciThrowException(&iRc, E_RECOV,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Null context provided", (char*)0);
    }
  
  compLog(0, "  contextAttribute: %p", (void*)contextAttr);
  compLog(0, "Searching current attribute for key: %s",
	  DmbString(key));
  if (!CciCharCmp(key, contextAttr->attributeName))
    {
      compLog(0, "Found value: %s", DmbString(contextAttr->attributeValue));
      returnvalue = contextAttr->attributeValue;
    }
  else
    {
      compLog(0, "Key not found in this attribute");
      if (contextAttr->next != NULL)
	{
	  compLog(0, "Trying next attribute structure in chain");
	  returnvalue=getContextAttribute(key, contextAttr->next);
	}
      else
	{
	  compLog(0, "No further attributes structures to try: key not found");
	  returnvalue = NULL;
	}
    }
  compLog(-1,functionName);
  return(returnvalue);
} 

/*  put attribute */

int __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX putContextAttribute(CciChar* in_key, CciChar* in_value, 
			struct ContextAttribute* contextAttr,
			int create)
{
  CciChar* key;
  CciChar* value;
  int iRc;
  static char* functionName = "putContextAttribute()";
  int returnValue;

  struct ContextAttribute* newOne;
  compLog(1,functionName);

  if (contextAttr==NULL)
    {
      cciThrowException(&iRc, E_RECOV,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Null context attribute provided", (char*)0);
    }

  compLog(0, "  contextAttribute: %p", (void*)contextAttr);

  /*  copy the data we're given so we can delete it with impunity */
  compLog(0, "putting key: %s, value: %s",
	  DmbString(in_key), DmbString(in_value));

  if (contextAttr->attributeName==NULL && create)
    {
      /*  we can store the data in this object */
      compLog(0, "Current attribute structure has no previous key "
	      "and create flag is set. "
	      "Storing here. Allocating memory to copy key and value");
      key=malloc(sizeof(CciChar)*CciCharLen(in_key)+1);
      TESTMALLOC(key);
      value=malloc(sizeof(CciChar)*CciCharLen(in_value)+1);
      TESTMALLOC(value);
      CciCharCpy(key, in_key);
      CciCharCpy(value, in_value);
      contextAttr->attributeName=key;
      contextAttr->attributeValue=value;
      returnValue=0;
    }
  /*  if we already have a key of this name, replace its value */
  else if (contextAttr->attributeName!=NULL && 
	   !CciCharCmp(contextAttr->attributeName, in_key))
  {
    compLog(0, "Key already exists; replacing its value");
    if (contextAttr->attributeValue!=NULL)
      {
	compLog(0, "Freeing storage occupied by previous value: %s",
		DmbString(contextAttr->attributeValue));
	free (contextAttr->attributeValue);
      }
    compLog(0, "Allocating storage for new value element");
    value=malloc(sizeof(CciChar)*CciCharLen(in_value)+1);
    TESTMALLOC(value);
    CciCharCpy(value, in_value);
    contextAttr->attributeValue=value;
    returnValue=0;
  }
  else if (contextAttr->next!=NULL)
    {
      /*  we didn't find a blank, we didn't match this one  */
      /*  and there are more to try: just recurse  */
      compLog(0, "Propagating request to next structure in chain");
      returnValue=putContextAttribute(in_key, in_value, contextAttr->next, create);
    }
  else if (contextAttr->next==NULL && create)
    {
      compLog(0, "Creating new attribute object to store this pair");
      /*  we must create a new context, link it to here */
      /*  and then recurse */
      newOne = (struct ContextAttribute*)malloc
	(sizeof (struct ContextAttribute));
      TESTMALLOC(newOne);
      initialiseContextAttribute(newOne);
      contextAttr->next=newOne;
      returnValue=putContextAttribute(in_key, in_value, newOne, create);
    }
  else 
  {
	/* We have not been asked to create an attribute and
	   there's none already defined with this name so
	   return 1
	*/
	  returnValue=1;
  }
  compLog(-1,functionName);
  return(returnValue);
}

/*  free attributes - recurse along next pointer and */
/*  then clear member elements.  */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX freeContextAttribute(struct ContextAttribute* contextAttr)
{
  static char* functionName="freeContextAttribute()";
  int iRc;
  compLog(1,functionName);

  if (contextAttr==NULL)
    {
      cciThrowException(&iRc, E_RECOV,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Null context provided", (char*)0);
    }
  compLog(0, "Context attribute at location %p", (void*)contextAttr);
  /*  free own variables */
  if(contextAttr->attributeName!=NULL)
    {
      free(contextAttr->attributeName);
    }
  if(contextAttr->attributeValue!=NULL)
    {
      compLog(0, "Freeing value %s.", 
	      DmbString(contextAttr->attributeValue));
      free(contextAttr->attributeValue);
    }
  /*  if we're not the last one, recurse */
  if(contextAttr->next!=NULL)
    {
      compLog(0, "Freeing next attribute...");
      freeContextAttribute(contextAttr->next);
    }
  /*  now delete ourself... */
  compLog(0, "Deleting ourself");
  free(contextAttr);

  compLog(-1,functionName);
}

/*  get attribute by index */

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX getContextAttributeIndex(struct 
				  ContextAttribute* contextAttr, 
				  int index)
{
  static char* functionName="getContextAttributeIndex()";
  int iRc;
  CciChar* returnValue;
  compLog(1,functionName);

  returnValue=NULL;
  
  if (contextAttr==NULL)
    {
      cciThrowException(&iRc, E_RECOV,RGB_CONTEXT_ATTRIBUTE_NULL,
			"Null context provided", (char*)0);
    }

  compLog(0, "ContextAttribute at: %p", (void*)contextAttr);
  compLog(0, "Requested name at position %d.", index);
  
  if (index==0)
    {
      compLog(0, "Found requested element. Name is %s",
	      DmbString(contextAttr->attributeName));
      returnValue=contextAttr->attributeName;
    }
  else
    {
      if (contextAttr->next==NULL)
	{
	  compLog(0, "Not found in this object and no more to search");
	  /*  out of bounds... */
	  returnValue=NULL;
	}
      else
	{
	  compLog(0, "Iterating to next attribute");
	  returnValue=getContextAttributeIndex(contextAttr->next, index-1);
	}
    }

  compLog(-1,functionName);
  return(returnValue);
}

/*  initialise context attribute */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX initialiseContextAttribute(struct ContextAttribute* contextAttr)
{
  static char* functionName="initialiseContextAttribute()";

  compLog(1,functionName);

  contextAttr->next=NULL;
  contextAttr->attributeName=NULL;
  contextAttr->attributeValue=NULL;

  compLog(-1,functionName);

  return;
}

/* //////////////////////////////////////////////////////////////////// */
/*  Message body traversal */
/*  */
/*  structure to represent syntax element -  */
/*  the idea is to have one of these objects for each element */
/*  Root.XML.blah.blah.blah would become five of these objects... */
/*  */
/*  The user provides the tree location in a free format field */
/*  Therefore, quotes should not be provided unless they are part */
/*  of the syntax element name */
/*  Not also that these structures are built up and torn down */
/*  for each message through. This is not terribly efficient. I'll */
/*  fix it in a later version */

/*  delete the mess afterwards... */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX deleteParseChain(struct parseChain* chain)
{
  static char* functionName="deleteParseChain()";

  int iRc;
  compLog(1,functionName);

  if (chain==NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_NULL_POINTER_EXCEPTION2,
			"Null parse chain provided",
			"Null pointer provided for a parse chain",
			(char*)0);
    }

  compLog(0, "Parse chain at %p.", (void*)chain);

  if (chain->parseElement!=NULL)
    {
      compLog(0, "Freeing parseElement %s", 
	      DmbString(chain->parseElement));
      free(chain->parseElement);
    }

  if (chain->next!=NULL)
    {
      compLog(0, "Propagating delete along chain");
      deleteParseChain(chain->next);
    }

  compLog(0,"Releasing storage for this structure");
  free(chain);
  
  compLog(-1,functionName);

  return;
}

/*  take . separated string, return CciChar* containing first */
/*  element and modify input pointer to point to next (null if */
/*  no more... */
/*  We do not modify the input string and the returned */
/*  strings should be freed (usually as part of deleteParseChain() ) */

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX getNextSyntaxElement(CciChar** currentPointer)
{
  static char* functionName="getNextSyntaxElement()";

  int iRc;
  CciChar* PERIOD;
  CciChar* returnValue; 
  CciChar* rVcursor;

  compLog(1,functionName);

  compLog(0, "Initialising string constants");

  PERIOD = CciString(".");

  if (currentPointer==NULL||*currentPointer==NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_NULL_POINTER_EXCEPTION2,
			"Null pointer for syntax element",
			"Null pointer for syntax element",
			(char*)0);
    }  

  compLog(0,"Pointer to CciString*: %p, CciString*: %p",
	  (void*)currentPointer, (void*)*currentPointer);

  compLog(0, "Input string is: %s.", DmbString(*currentPointer));

  /*  we don't know how large the next string to be extracted */
  /*  will be but we know it can't be any longer than the  */
  /*  whole string! */
  
  compLog(0,"Allocating space for returned portion"); 
  returnValue = (CciChar*)malloc((CciCharLen(*currentPointer)+1) *
				 sizeof(CciChar));
  TESTMALLOC(returnValue);

  /*  we don't want to change the value of the returnValue  */
  /*  pointer since we want */
  /*  to return it at the end of this! so use a copy... */

  compLog(0, "Searching for period or null");
  rVcursor=returnValue;
  
  /* Figure out what kind of leaps the compiler will make... */
  compLog(0, "size of   currentPointer: %d", sizeof(  currentPointer));
  compLog(0, "size of  *currentPointer: %d", sizeof( *currentPointer));
  compLog(0, "size of **currentPointer: %d", sizeof(**currentPointer));

  /* The following loop is deliberately verbose; *a++ does not
     necessarily parse the way one would expect :-)  */

  while((**currentPointer)!=((CciChar)0) && (**currentPointer)!=(*PERIOD))
    {
      compLog(0, "Found non period or null character; copying to output");
      *rVcursor=**currentPointer;
      rVcursor=rVcursor + 1;
      compLog(0, "*currentPointer is: %p", (void*)*currentPointer);
      *currentPointer = *currentPointer + 1;
      compLog(0, "*currentPointer has been incremented to: %p", 
	      (void*)*currentPointer);
    }

  /* Find out why we exited the loop */
  
  if (**currentPointer==((CciChar)0))
    {
      compLog(0, "Exited since **currentPointer was null");
    }

  /*  *currentPointer holds the address of the next . or the terminal */
  /*  null. If it is a . then we should advance  by one.  */
  /*  Note that if a user types two periods, this advancement ensures we */
  /*  don't loop forever - we just return empty - but well-formed strings */
  /*  It is the responsibility of the other elements of this node */
  /*  to ensure we treat empty strings correctly, however... */

  if (**currentPointer==*PERIOD)
    {
      compLog(0, "Found period; advancing past");
      *currentPointer=*currentPointer+1;
    }

  /*  we null-terminate the new string */
  compLog(0, "Null terminating new string");
  *rVcursor=0;

  compLog(0, "Returned string: %s; residual string: %s",
	  DmbString(returnValue), DmbString(*currentPointer));
  
  compLog(0, "Freeing string constants");
  free(PERIOD);
  
  compLog(-1,functionName);
  return returnValue;
}

/*  initialise parseChain */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX initialiseParseChain(struct parseChain* pC)
{
  static char* functionName="initialiseParseChain()";

  int iRc;
  compLog(1,functionName);

  if (pC==NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_NULL_POINTER_EXCEPTION2,
			"Null pointer for parse chain",
			"Null pointer for parse chain",
			(char*)0);
    }
  compLog(0, "ParseChain at %p.", (void*)pC);

  pC->parseElement=NULL;
  pC->next=NULL;

  compLog(-1,functionName);

  return;
}

/*  helper put-code for parseChain */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX putParseChainElement(struct parseChain* chain, CciChar* value)
{
  static char* functionName="putParseChainElement()";
  int iRc;
  struct parseChain* newNext;

  compLog(1,functionName);
  compLog(0, "Putting value %s in chain: %p",
	  DmbString(value), (void*)chain);

  if (chain==NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_NULL_POINTER_EXCEPTION2,
			"Null pointer for parse chain",
			"Null pointer for parse chain",
			(char*)0);
    }
  if ((*value)==((CciChar)0))
    {
      /* A null string can occur if the user types a double 
	 period, for example. Log the fact and return;
	 do not add an element */
      compLog(0, "value is null so not adding to parse chain");
    }
  else
    {
      if (chain->parseElement==NULL)
	{
	  compLog(0, "Current parseChain element empty so storing here");
	  chain->parseElement=value;
	}
      else if (chain->next==NULL)
	{
	  compLog(0, "No space in current parse chain and no 'next' element "
		  "so creating one");
	  newNext=(struct parseChain*)malloc(sizeof(struct parseChain));
	  TESTMALLOC(newNext);
	  initialiseParseChain(newNext);
	  chain->next=newNext;
	  putParseChainElement(newNext, value);
	}
      else
	{
	  compLog(0, "Propagating to next parseChain element");
	  putParseChainElement(chain->next, value);
	}
    }
  compLog(-1,functionName);

  return;
}

/*  split input string into chain - splitting on the dots */

struct parseChain __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX splitString(CciChar* parseString)
{
  static char* functionName="splitString()";

  int iRc;
  CciChar* internalCursor;
  struct parseChain* initialParseChain;

  compLog(1,functionName);

  if (parseString==NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_NULL_POINTER_EXCEPTION2,
			"Null pointer provided to string to be parsed",
			"Null pointer provided to string to be parsed",
			(char*)0);
    }

  compLog(0, "String to split: %s", DmbString(parseString));
  compLog(0, "Allocating space for first parseChain element");
  
  internalCursor=parseString;
  initialParseChain=    
    (struct parseChain*)malloc(sizeof(struct parseChain));

  TESTMALLOC(initialParseChain);
  initialiseParseChain(initialParseChain);

  compLog(0, "Iterating across periods in input string");
  while((*internalCursor)!=((CciChar)0))
    {
      CciChar* currentElement;
      compLog(0, "passing %s into getNextSyntaxElement",
	      DmbString(internalCursor));
      currentElement=getNextSyntaxElement(&internalCursor);
      compLog(0, "getNextSyntaxElement has changed input string "
	      "to %s and returned %s.",	DmbString(internalCursor),
	      DmbString(currentElement));
      putParseChainElement(initialParseChain, currentElement);
    }

  compLog(-1,functionName);
  return (initialParseChain);
}


/*  searchParseTree() navigates the message tree for the 
    element represented by the search string. The element
    is returned if it exists. If searchFlags are set to
    RGB_CREATE, the element (and the path to it) are created
    if it doesn't exist.

    All created elements are name elements except for the
    final element which is a name-value element. This usually
    works well but may not do what you expect with messages
    owned by the XML parser.

    The path syntax accepted is different to the regular
    ESQL syntax. Specifically, the only correlation name
    supported is "Body". That is: Body.XXX is parsed as if
    Root.*[LAST].XXX had been provided.

    No quoting is supported: quote marks are treated as literals.
    No escaping is supported: periods are always treated
      element separators
    No array syntax is supported: That is: Root.*[LAST] will fail
    unless there is a parser named "*[LAST]" installed.

    This function should be called with the full search    
    string - the results are undefined if called half way down, etc 
    That is: the first four characters of searchString should be
    'Root' or 'Body'. This is not, however, enforced.
*/

CciElement __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX searchParseTree(CciElement* startingPoint, 
			    CciChar* searchString,
			    int searchFlags)
{
  static char* functionName="searchParseTree()";
  struct parseChain* chain;
  struct parseChain* chainSaved;
  int iRc;
  int error=0;
  CciChar* specialCase;
  CciChar* searchName;
  CciElement* currentElement=startingPoint;
  CciElement* tempElement;

  compLog(1, functionName);

  compLog(0, "Preparing string constants");
  specialCase=CciString("Body");

  compLog(0, "Searching for path: %s from element at %p with flags %d.",
	  DmbString(searchString),
	  (void*)startingPoint,
	  searchFlags);

  chain=splitString(searchString);
  chainSaved=chain; /*  keep pointer so we can release the storage */

  /*  At all points, assuming no errors, the currentElement */
  /*  pointer should point to the name element specified in the */
  /*  current chain structure. If not, there is an error. */
  /*  If we are at the top level, check for the Root.*[LAST] */
  /*  / Body equivalence. If this fails and at all other places, */
  /*  we return NULL if RGB_PRESERVE is set. Else, we create the  */
  /*  required name-element */

  /*  Firstly, handle the irritating special case */

  /*  if chain pointer is "Body", */
  /*  get last child */

  compLog(0, "Checking special case: Body==Root.*[LAST]");

  if (!CciCharCmp(specialCase, chain->parseElement))
    {
      /*  replace currentElement with its last child */
      /*  and hope for the best :-) */
      tempElement=currentElement;
      compLog(0, "Navigating to last child");
      currentElement=cniLastChild(&iRc, tempElement);
      if (iRc!=CCI_SUCCESS)
	{
	  /*  some sort of serious error */
	  compLog(0, "Unable to locate last child of root.");
	  cciLog(&iRc,L_ERROR,RGB_TREE_TRAVERSAL_FAILURE,
		 "Failed to locate last child of root!",
		 "Last child of root",
		 (char*)0);
	  error=1;
	}
      else if (iRc==CCI_SUCCESS && currentElement==NULL)
	{
	  compLog(0, "Last child did not exist! Either there "
		  "is no last child of root (unlikely) or this "
		  "routine has been called in an improper context");
	  cciLog(&iRc,L_ERROR,RGB_TREE_TRAVERSAL_FAILURE,
		 "Failed to locate last child of root!",
		 "Last child of root",
		 (char*)0);
	  /*  element doesn't exist */
	  /*  this code should only be executed on the Root */
	  /*  element to get down to the last child (the body) */
	  /*  This element really should exist so this is an  */
	  /*  error */
	  error=1;
	}
      /*  else things are actually OK! */
    }

  compLog(0, "Iterating along parse chain");
  while(error==0 && chain->next!=NULL)
    {
      chain=chain->next;
      searchName=chain->parseElement;
      /*  search children for this element. */
      tempElement=currentElement;
      compLog(0, "Searching amongst children for element %s.",
	      DmbString(searchName));
      currentElement=cniSearchFirstChild(&iRc,
					 tempElement,
					 CCI_COMPARE_MODE_NAME,
					 CCI_ELEMENT_TYPE_NAME /*irrelevant*/,
					 searchName);

      if (iRc!=CCI_SUCCESS)
	{
	  char* tempValue=mbString(searchName);
	  compLog(0, "Failed to find element. Return code was: %d", iRc);
	  cciLog(&iRc,L_ERROR,RGB_TREE_TRAVERSAL_FAILURE,
		 "Failed to find required element as child of current "
		 "element.",tempValue, (char*)0);
	  free(tempValue);
	  error=1;
	  currentElement=NULL;
	  break;
	}
      else if (iRc==CCI_SUCCESS && currentElement==NULL)
	{
	  /*  not there... */
	  compLog(0, "Required element not present");
	  if (searchFlags==RGB_PRESERVE)
	    {
	      char* tempValue=mbString(searchName);
	      compLog(0, "Flags do not permit creation so we have "
		      "failed to navigate to the required element");
	      cciLog(&iRc,L_ERROR,RGB_TREE_TRAVERSAL_FAILURE,
		     "Missing syntax element and seach flags to not "
		     "permit dynamic creation", tempValue, (char*)0);
	      free(tempValue);
	      error=1;
	      break;
	    }
	  else if (searchFlags==RGB_CREATE)
	    {
	      compLog(0, "Creating required element");
	      /*  create element and add as last child of tempElement; */
	      currentElement=cniCreateElementAsLastChild(&iRc,
							 tempElement);
	      if (iRc!=CCI_SUCCESS)
		{
		  compLog(0, "Failed to create an element as last child");
		  cciLog(&iRc,L_ERROR,RGB_CHILD_ELEMENT_CREATE_FAILED,
			 "Failed to create an element as last child" ,
			 (char*)0);
		  error=1;
		  break;
		}
	      compLog(0, "Element created as last child");
	      /*  If there are further chain elements, */
	      /*  make this a name element. Otherwise, */
	      /*  we make it a name-value element */
	      if (chain->next == NULL)
		{
		  compLog(0, "No next element in chain so setting to "
			  "be a name-value element");
		  cniSetElementType(&iRc,
				    currentElement,
				    CCI_ELEMENT_TYPE_NAME_VALUE);
		}
	      else
		{
		  compLog(0, "Further elements in chain so setting to "
			  "be a name element");
		  cniSetElementType(&iRc,
				    currentElement,
				    CCI_ELEMENT_TYPE_NAME);
		}
		   
	      if (iRc!=CCI_SUCCESS)
		{
		  compLog(0, "Failed to set element type");
		  cciLog(&iRc,L_ERROR, RGB_FAILED_TO_SET_ELEMENT_TYPE,
			 "Failed to set type of newly created element",
			 (char*)0);
		  error=1;
		  break;
		}
	      compLog(0, "Setting name of element");
	      cniSetElementName(&iRc,
				currentElement,
				searchName);
	      if (iRc!=CCI_SUCCESS)
		{
		  compLog(0, "Failed to set name of element to %s",
			  DmbString(searchName));
		  cciLog(&iRc,L_ERROR,RGB_FAILED_TO_SET_ELEMENT_NAME,
			 "Failed to set element name", (char*)0);
		  error=1;
		  break;
		}
	      /*  if we reached here, we have inserted a new */
	      /*  element and set its name (and type). */
	      /*  currentElement points to it */
	      /*  Life is good. */
	    }
	  else
	    {
	      /*  unknown searchFlag */
	      /*  log */
	      compLog(0, "Unknown search flag %d", searchFlags);
	      error=1;
	      break;
	    }
	}
      else
	{
	  /*  we did find it. Cool! */
	}
    }; /*  matches while() loop */

  /*  We have broken out of the loop either because we reached */
  /*  the end of the chain and have found our element (error=0) */
  /*  or because we could not navigate successfully. */
  /*  (error=1). This is usually because the path did not exist */
  /*  and we were not asked to create it */
  
  /*  free the elements of the chain using our cunningly saved  */
  /*  pointer to the first one */

  compLog(0, "Tidying up parse chain");
  deleteParseChain(chainSaved);
  compLog(0, "Deleting string constants");
  free(specialCase);
  compLog(-1,functionName);
  
  return (error?NULL:currentElement);
}

/* The misleadingly named replaceWithNameValue function does the 
   following:

   Take a pointer to a name or name-value element.

   If a name element:
   
     Search for the first child value elements if any.
     Detach it, remembering its location in the tree
       (i.e. remember a sibling if any, otherwise remember parent)
     Add a new value element in its place
     Return pointer to this element
   
   If a name-value element:
     Get pointer to parent (or siblings)
     Remember name of n-v element
     Detach name-value element
     Add new element in the same place
     Return pointer to this new element

   If neither, throw an exception

   The purpose of this function is to overcome the following problem:

     The location of the output data already contains data from
     the input messsage - either because input and output 
     locations are identical or for some other reason.

     Who is responsible for freeing the storage corresponding to the
     existing data?

   The answer is that the broker is responsible. The easiest way to make
   this happen is to detach the element and replace with a blank one.
   The broker undertakes to garbage collect the storage for us.
   Whether it does or not is an entirely different matter...

*/

CciElement __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX replaceWithNameValue(CciElement* targetElement)
{
  static char* functionName = "replaceWithNameValue()";

  CciElementType elementType;

  CciElement* parent;
  CciElement* leftSibling;
  CciElement* newElement;

  CciChar* name;

  int iRc;
  int firstChild;

  compLog(1,functionName);
  
  compLog(0, "Allocating space to hold name of removed elements");

  name = malloc(sizeof(CciChar)*1024);
  TESTMALLOC(name);
  
  firstChild=0;
  newElement = targetElement;

  compLog(0, "Obtaining reference to parent of target element");

  parent = cniParent(&iRc, newElement);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to navigate to parent element",
			"Failed to navigate to parent element",
			(char*)0);
    } 

  compLog(0, "Reference to parent obtained. Address: %p.", 
	  (void*)parent);
  compLog(0, "Obtaining reference to left sibling of target element "
	  "(if one exists)");

  leftSibling = cniPreviousSibling(&iRc,
				   newElement);

  if (iRc != CCI_SUCCESS)
    {
      compLog(0, "No left sibling found");
      /* we assume this is because it has no previous sibling */
      firstChild=1;
    }
  else
    {
      compLog(0, "Left sibling located. Address is %p.", (void*)leftSibling);
    }

  compLog(0,"Obtaining name of target element");
  /* ignore CciSize - we just throw an 
     exception if buffer not large enough */
  cniElementName(&iRc, 
		 newElement,
		 name,
		 1023
		 );

  if (iRc != CCI_SUCCESS)
    {
      compLog(0, "Exception (%d) determining name of element. "
	      "Note that this could be"
	      "because the name is larger than the buffer", iRc);
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to determine name of element",
			"Failed to determine name of element",
			(char*)0);
    } 

  compLog(0,"Name is %s.", DmbString(name));
  compLog(0,"Determining type of target element");
  elementType = cniElementType(&iRc,
			       newElement);

  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to determine type of element",
			"Failed to determine type of element",
			(char*)0);
    } 

  /*  Note that element type should only be name or name-value */
  /*  however, we don't enforce this so things could go wrong... */
  /*  if name, navigate down to first value element (create if it doesn't */
  /*  exist - we will delete it very shortly, however) The reason */
  /*  for this apparent over-work is since I want to let the broker */
  /*  take full responsibility for deleting elements that I didn't */
  /*  necessarily create */

  if (elementType==CCI_ELEMENT_TYPE_NAME)
    {
      compLog(0, "Type determined to be NAME. "
	      "Locating first child value element. Note that we "
	      "ignore the possibility of more than one valid child "
	      "and concern ourselves with just the first");
      compLog(0, "Adjusting parent pointer to point to the name "
	      "element that owns the value element we are targetting. "
	      "Pointer was: %p. Becomes: %p.", (void*)parent,
	      (void*)newElement);
      parent=newElement;
      newElement = cniSearchFirstChild(&iRc,
				       newElement,
				       CCI_COMPARE_MODE_GENERIC_TYPE,
				       CCI_ELEMENT_TYPE_VALUE,
				       0);
      if (iRc != CCI_SUCCESS)
	{
	  /*  This is most likely because there is no value child. */
	  /*  So add one as first child! This will be deleted very soon */
	  /*  but for simplicity, we add it to save worrying about null */
	  /*  pointers and all that rubbish */

	  compLog(0, "No value elements as children. Creating a "
		  "temporary one");

	  newElement = cniCreateElementAsFirstChild(&iRc,
						    parent);
	  /*  we don't care about the return code since if there was */
	  /*  a problem, we will find out soon enough. */
	  
	  /*  Note: we set the type attribute here in case the detach */
	  /*  code requires it (which is not unlikely) */

	  compLog(0, "Temporarily setting type to value element ("
		  "This element will be deleted very soon so this "
		  "stage is simply to make later code simpler)");

	  cniSetElementType(&iRc,
			    newElement,
			    CCI_ELEMENT_TYPE_VALUE);
	  /*  Again, we don't care about the return code since we */
	  /*  will very shortly detach this element anyway... */
	  /*  Watch out for errors caused by this value element not */
	  /*  having been set */
	}

      firstChild = 0;
      /* parent already points to the name element since
	 we moved it down earlier on in this block*/

      compLog(0, "Recalculating leftSibling now we have navigated down "
	      "to the value element");

      leftSibling = cniPreviousSibling(&iRc,
				       newElement);
      
      if (iRc != CCI_SUCCESS)
	{
	  compLog(0, "No previous sibling so this element is the "
		  "first child of its parent");
	  /* we assume this is because it has no previous sibling */
	  firstChild=1;
	}
      else
	{
	  compLog(0, "Left sibling determined to be at %p.",
		  (void*)leftSibling);
	}
      
    }
  else
    {
      compLog(0, "Element not of type NAME. We do not check further. "
	      "Therefore, we assume it is a name-value element");
    }
		
  compLog(0, "Detaching target element (i.e. value portion of name element or "
	  "whole name-value element as appropriate");

  cniDetach(&iRc,
	    newElement);

  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to detach target element",
			"Failed to detach target element",
			(char*)0);
    } 

  if (!firstChild)
    {
      compLog(0, "Target element had a previous sibling so creating "
	      "replacement element as right sibling of previously saved "
	      "sibling (%p).", (void*)leftSibling);
      newElement = cniCreateElementAfter(&iRc,
					 leftSibling);
    }
  else
    {
      compLog(0, "Target element was first relevant child of its parent "
	      "so creating replacemenet as first child of root. This "
	      "(necessarily) can result in loss of ordering across types. "
	      "Parent previously calculated to be %p.", (void*)parent);
      newElement = cniCreateElementAsFirstChild(&iRc,
						parent);
    }
  
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to create child element",
			"Failed to create child element",
			(char*)0);
    } 

  compLog(0, "Setting type of element to match previously detached element");

  if (elementType == CCI_ELEMENT_TYPE_NAME_VALUE)
    {
      cniSetElementType(&iRc,
			newElement,
			CCI_ELEMENT_TYPE_NAME_VALUE);
    }
  else
    {
      /*  Then we must have detached a value element (i.e. */
      /*  target element type was a name element) */
      cniSetElementType(&iRc,
			newElement,
			CCI_ELEMENT_TYPE_VALUE);
    }
  
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			"Failed to set element type",
			"Failed to set element type",
			(char*)0);
    } 

  if (elementType == CCI_ELEMENT_TYPE_NAME_VALUE)
    {
      compLog(0, "Target element is a name-value element; setting name "
	      " to be %s.", DmbString(name));
      cniSetElementName(&iRc,
			newElement,
			name);
      
      if (iRc != CCI_SUCCESS)
	{
	  if (iRc == CCI_EXCEPTION)
	    {
	      cciRethrowLastException(&iRc);
	    }
	  cciThrowException(&iRc,E_RECOV,RGB_GENERAL_FAILURE,
			    "Failed to set element name",
			    "Failed to set element name",
			    (char*)0);
	}   
    }

  /*  This free call assumes the broker copies the name of */
  /*  newly created elements... */

  compLog(0, "ASSUMPTION: Broker copies the names of elements when their "
	  "names are set. On this assumption, freeing the name storage");

  free(name);

  compLog(-1,functionName);

  return(newElement);

}

/*  Take node context, input message and other goodies to produce */
/*  output message ready for data to be written */
/*  to. If the input and output locations are the same, we ensure the */
/*  destination is taken care of (i.e. the tree copy is followed by the */
/*  relevant detach... */
/*  */
/*  Upon return from this function, the inputElement and outputElement */
/*  members of the msgElts structure will */
/*  point to the correct value or name-value element.  */
/*  i.e. the caller can use the setvalue group without worrying about */
/*  the type. */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX prepareMessage(CciMessage* inputMessage,
		    struct CompDecompNodeContext* nodeContext,
		    struct messageElements* msgElts)
{
  static char* functionName = "prepareMessage()";


  static CciChar* bodyValue;   
  static CciChar* inputString; 
  static CciChar* outputString;

  CciMessageContext* messageContext;
  CciChar* inputLocation;
  CciChar* outputLocation;
  
  CciElement* inputRoot;
  CciElement* outputRoot;

  int iRc;
  
  compLog(1,functionName);

  compLog(0, "Initialising string constants");
  bodyValue = CciString("Root.BLOB.BLOB");
  inputString = CciString("inputSource");
  outputString = CciString("outputSource");  
  
  /*  get message context */
  compLog(0, "Requesting message context...");
  messageContext = cniGetMessageContext(&iRc, inputMessage);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_FAILED_GET_MESSAGE_CONTEXT,
			"Failed to get message context",
			(char*)0);
    }

  compLog(0, "Message context determined to be at: %p.", 
	  (void*)messageContext);

  /*  discover input location */
  compLog(0, "Determining location in syntax tree of input data");
  inputLocation = getContextAttribute(inputString, 
				      nodeContext->attribLinkedList);
  if (inputLocation == NULL)
    {
      compLog(0, "No value found for the %s key. Reverting to default: %s",
	      DmbString(inputString), DmbString(bodyValue));
      inputLocation = bodyValue;
    }

  /*  discover output location */
  compLog(0, "Determining location in syntax tree of output data");
  outputLocation = getContextAttribute(outputString, 
				       nodeContext->attribLinkedList);
  if (outputLocation == NULL)
    {
      compLog(0, "No value found for the %s key. Reverting to default: %s",
	      DmbString(outputString), DmbString(bodyValue));
      outputLocation = bodyValue;
    }

  /*  create output message */
  compLog(0, "Creating output message object");
  msgElts->outputMessage = cniCreateMessage(&iRc, messageContext);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_FAILED_CREATE_OUTBOUND_MSG,
			"Failed to create output message",
			(char*)0);
    }

  /*  Copy input tree to output tree */

  compLog(0, "Preparing to copy input tree to output tree: finding root of "
	  "input tree");

  inputRoot = cniRootElement(&iRc, inputMessage);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_FAILED_GET_ROOT,
			"Failed to locate root element of input message tree",
			(char*)0);
    }

  compLog(0, "Input tree root determined(%p); locating root of output tree.",
	  (void*)inputRoot);

  outputRoot = cniRootElement(&iRc, msgElts->outputMessage);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_FAILED_GET_ROOT,
			"Failed to locate root of output message tree.",
			(char*)0);
    }

  compLog(0, "Output tree root determined(%p); copy input tree to output "
	  "tree.", (void*)outputRoot);

  cniCopyElementTree(&iRc, inputRoot, outputRoot);
  if (iRc != CCI_SUCCESS)
    {
      if (iRc == CCI_EXCEPTION)
	{
	  cciRethrowLastException(&iRc);
	}
      cciThrowException(&iRc,E_RECOV,RGB_FAILED_COPY_MESSAGE_TREE,
			"Failed to copy input tree to output tree",
			(char*)0);
    }

  compLog(0, "Successful copying. Searching for pointer to input data");

  /*  find input element */
  msgElts->inputElement = searchParseTree(inputRoot, 
					  inputLocation,
					  RGB_PRESERVE);
  if (msgElts->inputElement == NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_TREE_TRAVERSAL_FAILURE,
			"Failed to locate input element in message tree",
			"Input element",
			(char*)0);
    }

  compLog(0, "Found input data(%p). Searching for pointer to output data",
	  (void*)msgElts->inputElement);

  /*  find output element */
  msgElts->outputElement = searchParseTree(outputRoot,
					   outputLocation,
					   RGB_CREATE);
  if (msgElts->outputElement == NULL)
    {
      cciThrowException(&iRc,E_RECOV,RGB_TREE_TRAVERSAL_FAILURE,
			"Failed to locate output element",
			"Output element",
			(char*)0);
    }

  compLog(0, "Found output data(%p). Now conditioning output element ",
	  (void*)msgElts->outputElement);

  /* Note that replaceWithNameValue() is a little more subtle than 
     its name might suggest: it merely replaces the smallest possible element
     it can with a fresh one (i.e. a name value is replaced with a new
     name-value and a value is replaced with a new value.

     This is done purely to let the broker take care of any data that
     might be present in these existing elements: we don't
     want any responsibility for freeing the storage: therefore, the
     detach call is required */

  msgElts->outputElement = replaceWithNameValue(msgElts->outputElement);

  compLog(0, "Freeing string constants");
  free(bodyValue);
  free(inputString);
  free(outputString);
  
  compLog(-1,functionName);

  return;
}
  
/* /////////////////////////////////////////////////////////// */
/* Adapted from mqsixLog in NodeUtils.c by Neil Kolban (with permission) */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX compLog(int offset, char *szFormat, ...)
{
#ifndef _DEBUG
  return;
#else
  
  /* This function has been modified to allow indenting for 
     function calls. The current indent is stored in the static
     integer printOffset.

     This will only work as expected when the function is only
     accessible to one thread: The lack of synchronisation
     and sharing of the variable across threads means that integrity
     cannot be assured in multiple compressor/decompressor node
     environments. 

     Accordingly, the _DEBUG variable should be set only for
     testing in a single-threaded environment

  */

  static FILE	*logFile = NULL;
  static int    printOffset=0;
  struct tm	*now_tm;
  time_t	iNow;
  va_list	list;
  int           count;

  va_start(list, szFormat);
  
  printOffset += offset;

  if (logFile == NULL)
    {
      logFile = fopen("__RGB_DEBUG_OUTPUT_LOCATION", "a");
      if (logFile == NULL)
	{
	  return;
	}
    }
  
  iNow = time(NULL);
  now_tm = localtime(&iNow);
  fprintf(logFile, "%.2d/%.2d/%.4d %.2d:%.2d:%.2d ",
	  now_tm->tm_mon+1,
	  now_tm->tm_mday,
	  now_tm->tm_year+1900,
	  now_tm->tm_hour,
	  now_tm->tm_min,
	  now_tm->tm_sec);

  count=1; /* initial function will indent us one but this
	      is unnecessary */

  while(count<=printOffset)
    {
      fprintf(logFile, "  ");
      count+=1;
    }

  if (offset>0)
    {
      fprintf(logFile, "> ");
      /*count+=1;*/ /* this is necessary to overcome the added two spaces
		   we have just created */
    }
  else if (offset<0)
    {
      fprintf(logFile, "  < ");  /* We have already moved back two places by this stage... */
      /*count-=1;*/
    }
  else
  {
	  fprintf(logFile, "  ");
  }

  vfprintf(logFile, szFormat, list);
  fprintf(logFile, "\n");
  fflush(logFile);
  va_end(list);
#endif
}

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX initMessageSourcePointer()
{
	messageSourcePointer = CciString("CompressorCatalogue");
}
