/*  Richard G Brown */
/*  Compression / Decompression plug-in */
/*  nodes for MQSeries Integrator version 2 */
/*  Helper utilities */

#ifndef __RGB_NODE_UTILS
#define __RGB_NODE_UTILS

/* The following #include allows rgbnodeutils.c to define __RGB_IMPORTEXPORT
   as _declspec(dllexport) (for Windows at least) whilst USERS of the
   library get the import version.
*/

#include "rgbos.h"

#include <malloc.h>
#include <memory.h>
#include <stdarg.h>
#include <wchar.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <BipCci.h>    /*  Common functions (between nodes and parsers) */
#include <BipCni.h>    /*  Message processing nodes */

/* define _DEBUG to produce copious trace.
   This tracing code is NOT THREAD SAFE.
   
   Therefore, only apply trace to message flows containing ONE message flow
   and ONE node

   It doesn't even work with two nodes in the same flow!
*/

/*#define _DEBUG*/

#include "compressorcat.h"

extern CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX messageSourcePointer;  /* Defined in rgbnodeutils.c */

/* //////////////////////////////////////////////////////////////////// */
/* Support for logging and exception handling; these constants take the 
   place of the five standard entries in the CciLog and CciThrowException
   calls */

/* Note that if CciString() fails on startup, messageSourcePointer will be null
   and any logging caused by the failure of CciString() will be compromised.
   This needs looking addressing in a later version */

#define RGB_PRESERVE 0
#define RGB_CREATE 1

#define COMPRESSOR_MESSAGE_SOURCE messageSourcePointer

/* A few macros to make logging slightly less painful */

#define L_INFO CCI_LOG_INFORMATION,__FILE__,__LINE__,functionName,COMPRESSOR_MESSAGE_SOURCE
#define L_ERROR CCI_LOG_ERROR,__FILE__,__LINE__,functionName,COMPRESSOR_MESSAGE_SOURCE
#define E_FATAL CCI_FATAL_EXCEPTION,__FILE__,__LINE__,functionName,COMPRESSOR_MESSAGE_SOURCE
#define E_CONFIG CCI_CONFIGURATION_EXCEPTION,__FILE__,__LINE__,functionName,COMPRESSOR_MESSAGE_SOURCE
#define E_RECOV CCI_RECOVERABLE_EXCEPTION,__FILE__,__LINE__,functionName,COMPRESSOR_MESSAGE_SOURCE

/* This is probably overkill and would have been better implemented by replacing malloc... */

#define TESTMALLOC(MALLOCRC) if(MALLOCRC==0){int _EMERG; compLog(0, "Malloc failure"); cciThrowException(&_EMERG,CCI_FATAL_EXCEPTION,__FILE__, __LINE__,functionName, COMPRESSOR_MESSAGE_SOURCE,RGB_MALLOC_FAILURE,"Malloc failure", (char*)0);};

/* //////////////////////////////////////////////////////////////////// */
/*  node context definition */

/* //////////////////////////////////////////////////////////////////// */
/*  Context Handling */

/*  attribute definition */

struct ContextAttribute
{
  struct ContextAttribute* next;
  CciChar* attributeName;
  CciChar* attributeValue;
};


struct CompDecompNodeContext
{
  /*  Context Attributes linked list */
  struct ContextAttribute* attribLinkedList;

  /*  Required fields for terminals */
  CciTerminal* outTerminal;
  CciTerminal* failureTerminal;
  CciTerminal* inTerminal;
};


/* //////////////////////////////////////////////////////////////////// */
/*  String handling functions from sample plug in code */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX CciCharCpy(CciChar* target, const CciChar* source);

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX CciCharNCpy(CciChar* target, const CciChar* source, 
		  int size);

int __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX CciCharLen(const CciChar* s);

int __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX CciCharCmp(const CciChar * s1, const CciChar* s2);

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX CciString(const char* source);

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX CciNString(const char* source, const unsigned int size);

char __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX mbString(const CciChar* source);

char __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX DmbString(const CciChar* source);


/*  get attribute */

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX getContextAttribute(CciChar* key, 
			     struct ContextAttribute* contextAttr);

/*  put attribute
    create flag: 1 add an extra attribute if not already present; return 0
    create flag: 0 do not add the attribute if it doesn't exist
                   return 0 if it existed; 1 if it did not
 */

int __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX putContextAttribute(CciChar* in_key, CciChar* in_value, 
			 struct ContextAttribute* contextAttr, 
			 int create);

/*  free attributes - recurse along next pointer and */
/*  then clear member elements.  */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX freeContextAttribute(struct ContextAttribute* contextAttr);

/*  get attribute by index */

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX getContextAttributeIndex(struct 
				  ContextAttribute* contextAttr, 
				  int index);

/*  initialise context attribute */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX 
	initialiseContextAttribute(struct ContextAttribute* contextAttr);


/* //////////////////////////////////////////////////////////////////// */
/*  Message body traversal */

/*  structure to represent syntax element -  */
/*  the idea is to have one of these objects for each element */
/*  Root.XML.blah.blah.blah would become five of these objects... */
/*  I don't know if the broker passes on quotes if the user is */
/*  annoying enough to specify them. We check, just in case, however. */
/*  of course, if quotes are part of the name then we are in trouble */

struct parseChain
{
  CciChar* parseElement;
  struct parseChain* next;
};

struct messageElements
{
  CciMessage* outputMessage;
  CciElement* inputElement;
  CciElement* outputElement;
};

/*  delete the mess afterwards... */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX deleteParseChain(struct parseChain* chain);

/*  take . separated string, return CciChar* containing first */
/*  element and modify input pointer to point to next (null if */
/*  no more... */
/*  We do not modify the input string and the returned */
/*  strings should be free'd (usually as part of deleteParseChain) */

CciChar __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX getNextSyntaxElement(CciChar** currentPointer);

/*  initialise parseChain */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX initialiseParseChain(struct parseChain* pC);

/*  helper put-code for parseChain */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX putParseChainElement(struct parseChain* chain, CciChar* value);

/*  split input string into chain - splitting on the dots */

struct parseChain __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX splitString(CciChar* parseString);

/* See rgbnodeutils.c for a full discussion of this function */

#define RGB_PRESERVE 0
#define RGB_CREATE 1

CciElement __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX searchParseTree(CciElement* startingPoint, 
			    CciChar* searchString,
			    int searchFlags);

/* See rgbnodeutils.c for a full discussion of this function */

CciElement __RGB_IMPORTEXPORTPREFIX * __RGB_IMPORTEXPORTSUFFIX replaceWithNameValue(CciElement* targetElement);

/* See rgbnodeutils.c for a full discussion of this function */

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX prepareMessage(CciMessage* inputMessage,
		    struct CompDecompNodeContext* nodeContext,
		    struct messageElements* msgElts);

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX compLog(int offset, char *szFormat, ...);

void __RGB_IMPORTEXPORTPREFIX __RGB_IMPORTEXPORTSUFFIX initMessageSourcePointer();

#endif

